#pragma once

#include "exc_return.h"

#define CONTROL_NPRIV (UINT32_C(1) << 0)

#define PSR_THUMB (UINT32_C(1) << 24)

#define DWT_LAR (*(volatile uint32_t *)0xE0001FB0)
#define DWT_LAR_UNLOCK UINT32_C(0xC5ACCE55)

#define DEMCR (*(volatile unsigned *)0xE000EDFCU)
#define DEMCR_TRCENA (UINT32_C(1) << 24)

#define DWT_CTRL (*(volatile uint32_t *)0xE0001000U)
#define DWT_CTRL_CYCCNTENA (UINT32_C(1) << 0)

#define DWT_CYCCNT (*(volatile uint32_t *)0xE0001004U)

struct context
{
#if RT_MPU_ENABLE
    /* In M-profile, thread-mode privilege is part of the control register. In
     * A/R-profile, it's part of the processor mode (system vs. user), which is
     * a field of the CPSR. */
    uint32_t control;
#endif

#if V8M
    uint32_t psplim;
#endif

    uint32_t r4, r5, r6, r7, r8, r9, r10, r11;

#if FPU
    /* Only use a per-task exception return value if floating-point is enabled,
     * because otherwise the exception return value is always the same. This
     * is the lr value on exception entry, so place it after r4-r11 so it can
     * be saved/restored along with those registers. */
    uint32_t exc_return;
#endif

    uint32_t r0, r1, r2, r3, r12, lr, pc, psr;
};

static struct context *context_init(void *stack, size_t stack_size,
                                    uintptr_t fn_addr)
{
    (void)fn_addr;

    void *const sp = (char *)stack + stack_size;
    struct context *ctx = sp;
    --ctx;

#if RT_MPU_ENABLE
    /* Tasks start privileged. The SPSEL bit is RAZ/WI in handler mode where
     * context switches occur. The exception return value specifies which stack
     * pointer is used when returning to thread mode. */
    ctx->control = 0;
#endif

#if V8M
    ctx->psplim = (uint32_t)stack;
#endif

#if FPU
    ctx->exc_return = (uint32_t)TASK_INITIAL_EXC_RETURN;
#endif

    ctx->psr = PSR_THUMB;

    return ctx;
}

void rt_task_drop_privilege(void)
{
#if (__ARM_ARCH >= 7) && RT_MPU_ENABLE
    /* We only track per-task privilege in M-profile if the MPU is enabled.
     * Also, dropping privilege on v6-m prevents the use of atomic operations,
     * which require interrupt masking on that architecture. */
    uint32_t control;
    __asm__ __volatile__("mrs %0, control" : "=r"(control));
    __asm__("dsb; msr control, %0; isb"
            :
            : "r"(control | CONTROL_NPRIV)
            : "memory");
#endif
}

void rt_cycle_init(void)
{
#if (__ARM_ARCH >= 7)
    // Enable the cycle counter.
    DWT_LAR = DWT_LAR_UNLOCK;
    DEMCR |= DEMCR_TRCENA;
    DWT_CTRL |= DWT_CTRL_CYCCNTENA;
#endif
}

bool rt_interrupt_is_active(void)
{
    uint32_t ipsr;
    __asm__ __volatile__("mrs %0, ipsr" : "=r"(ipsr));
    return ipsr != 0;
}

#define ICSR (*(volatile uint32_t *)0xE000ED04UL)
#define PENDSVSET (UINT32_C(1) << 28)

void rt_syscall_pend(void)
{
    ICSR = PENDSVSET;
}

uint32_t rt_cycle(void)
{
#if RT_CYCLE_ENABLE && (__ARM_ARCH >= 7)
    return DWT_CYCCNT;
#else
    return 0;
#endif
}
