#pragma once

#include "coprocessor.h"
#include "mode.h"

#include <vic.h>

#define PSR_THUMB (UINT32_C(1) << 5)
#define PSR_E (UINT32_C(1) << 9)

#define FPEXC_EN (UINT32_C(1) << 30)

#define BIG_ENDIAN (__BYTE_ORDER__ == __ORDER_BIG_ENDIAN__)

struct context
{
#if FPU
    /* CPACR controls whether floating-point is enabled. Give each task its own
     * copy of this register where floating-point is initially disabled, so
     * attempts to use floating-point in tasks cause undefined instruction
     * exceptions that can be handled by enabling floating-point and
     * re-executing the instruction. See section B1.11.3 of the Arm
     * Architecture Reference Manual ARMv7-A and ARMv7-R Edition. */
    uint32_t cpacr;
#endif

    uint32_t r4, r5, r6, r7, r8, r9, r10, r11;
    uint32_t r0, r1, r2, r3, r12, lr, pc, psr;
};

static struct context *context_init(void *stack, size_t stack_size,
                                    uintptr_t fn_addr)
{
    void *const sp = (char *)stack + stack_size;
    struct context *ctx = sp;
    --ctx;
    ctx->psr = MODE_SYS | (BIG_ENDIAN ? PSR_E : 0) |
               (((fn_addr & 1) != 0) ? PSR_THUMB : 0);
#if FPU
    ctx->cpacr = 0;
#endif
    return ctx;
}

void rt_task_drop_privilege(void)
{
    __asm__("dsb; cps %0" : : "i"(MODE_USR) : "memory");
}

void rt_cycle_init(void)
{
    // Enable counters and reset the cycle counter.
    pmcr_oreq(PMCR_E | PMCR_C);
    // Enable the cycle counter.
    pmcntenset_oreq(PMCNTEN_C);
}

static inline uint32_t cpsr_mode(void)
{
    uint32_t cpsr;
    __asm__ __volatile__("mrs %0, cpsr" : "=r"(cpsr));
    return cpsr & MODE_MASK;
}

static inline uint32_t spsr_mode(void)
{
    uint32_t spsr;
    __asm__ __volatile__("mrs %0, spsr" : "=r"(spsr));
    return spsr & MODE_MASK;
}

static inline bool mode_is_interrupt(uint32_t mode)
{
    /* NOTE: this assumes that nested interrupts don't use system mode.
     * Interrupt nesting should use supervisor mode, which doesn't require each
     * task stack to accommodate interrupts. */
    return (mode != MODE_USR) && (mode != MODE_SYS);
}

bool rt_interrupt_is_active(void)
{
    return mode_is_interrupt(cpsr_mode());
}

void rt_syscall_pend(void)
{
    vic_syscall_pend();
}

uint32_t rt_cycle(void)
{
#if RT_CYCLE_ENABLE
    return pmccntr();
#else
    return 0;
#endif
}

#if FPU
bool rt_lazy_enable_fp(void)
{
    /* If the undefined instruction was in an interrupt or the FPU is already
     * enabled, there's nothing to do. */
    if (mode_is_interrupt(spsr_mode()) ||
        ((cpacr() & CPACR_FPU_ENABLE) == CPACR_FPU_ENABLE))
    {
        return false;
    }

    // Enable the FPU and initialize the floating point state.
    cpacr_oreq(CPACR_FPU_ENABLE);
    __asm__("vmsr fpexc, %0" : : "r"(FPEXC_EN));
    __asm__("vmsr fpscr, %0" : : "r"(0));
    __asm__("vmov d0, %0, %0" : : "r"(0));
    __asm__("vmov d1, %0, %0" : : "r"(0));
    __asm__("vmov d2, %0, %0" : : "r"(0));
    __asm__("vmov d3, %0, %0" : : "r"(0));
    __asm__("vmov d4, %0, %0" : : "r"(0));
    __asm__("vmov d5, %0, %0" : : "r"(0));
    __asm__("vmov d6, %0, %0" : : "r"(0));
    __asm__("vmov d7, %0, %0" : : "r"(0));
    __asm__("vmov d8, %0, %0" : : "r"(0));
    __asm__("vmov d9, %0, %0" : : "r"(0));
    __asm__("vmov d10, %0, %0" : : "r"(0));
    __asm__("vmov d11, %0, %0" : : "r"(0));
    __asm__("vmov d12, %0, %0" : : "r"(0));
    __asm__("vmov d13, %0, %0" : : "r"(0));
    __asm__("vmov d14, %0, %0" : : "r"(0));
    __asm__("vmov d15, %0, %0" : : "r"(0));
    return true;
}
#endif
