#pragma once

#ifdef __cplusplus
extern "C" {
#endif

enum rt_syscall_op
{
    // Processes a tick.
    RT_SYSCALL_TICK,

    // Sleep from a task.
    RT_SYSCALL_TASK_SLEEP,
    RT_SYSCALL_TASK_SLEEP_PERIODIC,

    // Wait on a semaphore from a task.
    RT_SYSCALL_SEM_WAIT,
    RT_SYSCALL_SEM_TIMEDWAIT,

    // Post a semaphore from a task or interrupt.
    RT_SYSCALL_SEM_POST,

    // Lock a mutex from a task.
    RT_SYSCALL_MUTEX_LOCK,
    RT_SYSCALL_MUTEX_TIMEDLOCK,

    // Unlock a mutex from a task.
    RT_SYSCALL_MUTEX_UNLOCK,

    // Add a task to the ready list.
    RT_SYSCALL_TASK_READY,

    // Exit from a task.
    RT_SYSCALL_TASK_EXIT,
};

struct rt_sem;
struct rt_mutex;
struct rt_task;

union rt_syscall_args
{
    struct rt_syscall_args_task_sleep
    {
        unsigned long ticks;
    } task_sleep;
    struct rt_syscall_args_task_sleep_periodic
    {
        unsigned long last_wake_tick;
        unsigned long period;
    } task_sleep_periodic;
    struct rt_syscall_args_sem_wait
    {
        struct rt_sem *sem;
    } sem_wait;
    struct rt_syscall_args_sem_timedwait
    {
        struct rt_sem *sem;
        unsigned long ticks;
    } sem_timedwait;
    struct rt_syscall_args_sem_post
    {
        struct rt_sem *sem;
        int n;
    } sem_post;
    struct rt_syscall_args_mutex_lock
    {
        struct rt_mutex *mutex;
    } mutex_lock;
    struct rt_syscall_args_mutex_timedlock
    {
        struct rt_mutex *mutex;
        unsigned long ticks;
    } mutex_timedlock;
    struct rt_syscall_args_mutex_unlock
    {
        struct rt_mutex *mutex;
    } mutex_unlock;
    struct rt_syscall_args_task_ready
    {
        struct rt_task *task;
    } task_ready;
};

struct rt_syscall_record
{
    struct rt_syscall_record *next;
    union rt_syscall_args args;
    enum rt_syscall_op op;
};

/* Enqueue a system call record. It will be processed once the system call
 * handler is triggered, either by rt_syscall or rt_syscall_pend. */
void rt_syscall_push(struct rt_syscall_record *record);

/* Architecture-dependent trigger for the syscall handler to be called from
 * tasks. This should cause the syscall handler to run before this function
 * returns. */
void rt_syscall(void);

/* Architecture-dependent trigger for the syscall handler that can only be
 * called from an interrupt. If called from an interrupt, the syscall should
 * run once no other interrupts are running, but before another task gets to
 * run. */
void rt_syscall_pend(void);

/* Perform all pending system calls and return a new context to execute or NULL
 * if no context switch is required. */
void *rt_syscall_run(void);

/* Architecture-dependent handler for syscalls. This will call rt_syscall_run
 * and perform a context switch if necessary. */
void rt_syscall_handler(void);

#ifdef __cplusplus
}
#endif
