use core::{
    cell::UnsafeCell,
    marker::PhantomPinned,
    panic::{RefUnwindSafe, UnwindSafe},
};

use crate::{
    bindings::{rt_cond, rt_cond_broadcast, rt_cond_signal, rt_cond_timedwait, rt_cond_wait},
    sync::{semaphore::c_sem_init, MutexGuard},
    tick::Utick,
};

pub struct Condvar {
    c: UnsafeCell<rt_cond>,
    _pin_marker: PhantomPinned,
}

unsafe impl Send for Condvar {}
unsafe impl Sync for Condvar {}
impl UnwindSafe for Condvar {}
impl RefUnwindSafe for Condvar {}

impl Condvar {
    pub const fn init(cond: &'static Condvar) -> Condvar {
        let c = unsafe { &*(cond.c.get() as *const rt_cond) };
        Condvar {
            c: UnsafeCell::new(rt_cond {
                sem: c_sem_init(&c.sem, 0, 0),
            }),
            _pin_marker: PhantomPinned,
        }
    }

    /// Wake at most one task that is waiting on the `Condvar`.
    pub fn signal(&self) {
        unsafe { rt_cond_signal(self.c.get()) }
    }

    /// Wake all tasks that are waiting on the `Condvar`.
    pub fn broadcast(&self) {
        unsafe { rt_cond_broadcast(self.c.get()) }
    }

    /// Alias for `Condvar::signal` to match `std::sync::Condvar`.
    pub fn notify_one(&self) {
        self.signal()
    }

    /// Alias for `Condvar::broadcast` to match `std::sync::Condvar`.
    pub fn notify_all(&self) {
        self.broadcast()
    }

    /// Unlock the mutex until the `Condvar` has been signaled, and re-acquire the mutex afterward.
    pub fn wait<T>(&self, guard: &MutexGuard<T>) {
        unsafe { rt_cond_wait(self.c.get(), guard.ptr()) }
    }

    /// Repeatedly `wait` on the `Condvar` while `condition` returns `true`.
    pub fn wait_while<T, F>(&self, guard: &MutexGuard<T>, condition: F)
    where
        F: Fn(&T) -> bool,
    {
        while condition(guard) {
            self.wait(guard);
        }
    }

    /// `wait_while` but allow mutation of the guard and other state from the `condition` function.
    pub fn wait_while_mut<T, F>(&self, guard: &mut MutexGuard<T>, mut condition: F)
    where
        F: FnMut(&mut T) -> bool,
    {
        while condition(guard) {
            self.wait(guard);
        }
    }

    pub fn timed_wait<'a, T>(
        &self,
        guard: MutexGuard<'a, T>,
        ticks: Utick,
    ) -> Option<MutexGuard<'a, T>> {
        unsafe { rt_cond_timedwait(self.c.get(), guard.ptr(), ticks) }.then_some(guard)
    }
}

#[macro_export]
macro_rules! condvar {
    ($name: ident) => {
        static $name: $crate::sync::Condvar = $crate::sync::Condvar::init(&$name);
    };
}

#[cfg(test)]
mod tests {
    #[test]
    fn signal() {
        condvar!(CONDVAR);
        CONDVAR.signal();
    }
}
