#include <rt/mutex.h>

#include <rt/assert.h>
#include <rt/log.h>
#include <rt/syscall.h>
#include <rt/task.h>

#include <stddef.h>

void rt_mutex_init(struct rt_mutex *mutex)
{
    rt_list_init(&mutex->wait_list);
    rt_list_init(&mutex->list);
    rt_atomic_store(&mutex->holder, 0, RT_ATOMIC_RELEASE);
}

void rt_mutex_lock(struct rt_mutex *mutex)
{
    if (rt_mutex_trylock(mutex))
    {
        return;
    }

    rt_logf("%s mutex lock blocked\n", rt_task_name());
    struct rt_syscall_record lock_record;
    lock_record.args.mutex_lock.mutex = mutex;
    lock_record.op = RT_SYSCALL_MUTEX_LOCK;
    rt_syscall_push(&lock_record);
    rt_syscall();
}

bool rt_mutex_trylock_with_task(struct rt_mutex *mutex, struct rt_task *task)
{
    uintptr_t e = 0;
    const uintptr_t taskptr = (uintptr_t)task;
    if (rt_atomic_compare_exchange(&mutex->holder, &e, taskptr,
                                   RT_ATOMIC_ACQUIRE, RT_ATOMIC_RELAXED))
    {
        rt_logf("%s mutex lock success\n", task->name);
        return true;
    }
    else if ((e & RT_MUTEX_HOLDER_MASK) == taskptr)
    {
        // Already holding this mutex.
        rt_assert(mutex->level >= 0,
                  "locked a non-recursive mutex more than once");
        ++mutex->level;
        return true;
    }
    return false;
}

bool rt_mutex_trylock(struct rt_mutex *mutex)
{
    return rt_mutex_trylock_with_task(mutex, rt_task_self());
}

bool rt_mutex_timedlock(struct rt_mutex *mutex, unsigned long ticks)
{
    if (rt_mutex_trylock(mutex))
    {
        return true;
    }
    rt_logf("%s mutex timedlock blocked, timeout = %lu\n", rt_task_name(),
            ticks);

    struct rt_syscall_record lock_record;
    lock_record.args.mutex_timedlock.mutex = mutex;
    lock_record.args.mutex_timedlock.ticks = ticks;
    lock_record.op = RT_SYSCALL_MUTEX_TIMEDLOCK;
    rt_syscall_push(&lock_record);
    rt_syscall();
    return lock_record.args.mutex_timedlock.mutex != NULL;
}

void rt_mutex_unlock(struct rt_mutex *mutex)
{
    if (mutex->level > 0)
    {
        // Mutex is recursive and has been locked more than once.
        --mutex->level;
        return;
    }

    uintptr_t e = (uintptr_t)rt_task_self();
    if (rt_atomic_compare_exchange(&mutex->holder, &e, 0, RT_ATOMIC_RELEASE,
                                   RT_ATOMIC_RELAXED))
    {
        rt_logf("%s mutex unlock with no waiters\n", rt_task_name());
        // Unlock with no waiters, nothing to do.
        return;
    }

    /* This path should only be taken by a task, as any interrupt that unlocks
     * must have done a successful trylock, and there should never be any
     * waiting tasks, since an interrupt has been active for the entire
     * duration that the mutex was held. The active task pointer is only used
     * as a sentinel for the interrupt holding the mutex in this case. */
    rt_logf("%s mutex unlock\n", rt_task_name());
    struct rt_syscall_record unlock_record;
    unlock_record.args.mutex_unlock.mutex = mutex;
    unlock_record.op = RT_SYSCALL_MUTEX_UNLOCK;
    rt_syscall_push(&unlock_record);
    rt_syscall();
}
